
"""
"""


def lexer_scan(extensions):
    """ Scans extensions for ``lexer_rules`` and ``preprocessors``
        attributes.
    """
    lexer_rules = {}
    preprocessors = []
    postprocessors = []
    for extension in extensions:
        if hasattr(extension, 'lexer_rules'):
            lexer_rules.update(extension.lexer_rules)
        if hasattr(extension, 'preprocessors'):
            preprocessors.extend(extension.preprocessors)
        if hasattr(extension, 'postprocessors'):
            postprocessors.extend(extension.postprocessors)
    return {
        'lexer_rules': [lexer_rules[k] for k in sorted(lexer_rules.keys())],
        'preprocessors': preprocessors,
        'postprocessors': postprocessors
    }


class Lexer(object):
    """ Tokenizes input source per rules supplied.
    """

    def __init__(self, lexer_rules, preprocessors=None, postprocessors=None,
                 **ignore):
        """ Initializes with ``rules``. Rules must be a list of
            two elements tuple: ``(regex, tokenizer)`` where
            tokenizer if a callable of the following contract::

            def tokenizer(match):
                return end_index, token, value
        """
        self.rules = lexer_rules
        self.preprocessors = preprocessors or []
        self.postprocessors = postprocessors or []

    def tokenize(self, source):
        """ Translates ``source`` accoring to lexer rules into
            an iteratable of tokens.
        """
        for preprocessor in self.preprocessors:
            source = preprocessor(source)
        tokens = []
        append = tokens.append
        pos = 0
        lineno = 1
        end = len(source)
        while pos < end:
            for regex, tokenizer in self.rules:
                m = regex.match(source, pos, end)
                if m:
                    npos, token, value = tokenizer(m)
                    assert npos > pos
                    append((lineno, token, value))
                    lineno += source[pos:npos].count('\n')
                    pos = npos
                    break
            else:
                assert False, 'Lexer pattern mismatch.'
        for postprocessor in self.postprocessors:
            postprocessor(tokens)
        return tokens




"""
"""

from wheezy.template.comp import allocate_lock
from wheezy.template.loader import ChainLoader
from wheezy.template.loader import DictLoader


class Preprocessor(object):
    """ Preprocess templates with ``engine`` and vary runtime templates
        by ``key_factory`` function using ``runtime_engine_factory``.
    """

    def __init__(self, runtime_engine_factory, engine, key_factory):
        self.lock = allocate_lock()
        self.runtime_engines = {}
        self.runtime_engine_factory = runtime_engine_factory
        self.engine = engine
        self.loader = engine.loader
        self.key_factory = key_factory
        template_class = self.engine.template_class
        self.engine.template_class = lambda name, render_template: \
            template_class(name, lambda ctx, local_defs, super_defs:
                           self.render(name, ctx, local_defs, super_defs))

    def get_template(self, name):
        return self.engine.get_template(name)

    def render(self, name, ctx, local_defs, super_defs):
        try:
            runtime_engine = self.runtime_engines[
                self.key_factory(ctx)]
        except KeyError:
            runtime_engine = self.ensure_runtime_engine(
                self.key_factory(ctx))
        try:
            return runtime_engine.renders[name](ctx, local_defs, super_defs)
        except KeyError:
            self.preprocess_template(runtime_engine, name, ctx)
            return runtime_engine.renders[name](ctx, local_defs, super_defs)

    def remove(self, name):
        self.lock.acquire(1)
        try:
            self.engine.remove(name)
            for runtime_engine in self.runtime_engines.values():
                runtime_engine.remove(name)
        finally:
            self.lock.release()

    # region: internal details

    def ensure_runtime_engine(self, key):
        self.lock.acquire(1)
        try:
            engines = self.runtime_engines
            if key in engines:  # pragma: nocover
                return engines[key]
            engine = engines[key] = self.runtime_engine_factory(
                loader=ChainLoader([DictLoader({}), self.engine.loader]))

            def render(name, ctx, local_defs, super_defs):
                try:
                    return engine.renders[name](ctx, local_defs, super_defs)
                except KeyError:
                    self.preprocess_template(engine, name, ctx)
                    return engine.renders[name](ctx, local_defs, super_defs)
            engine.global_vars['_r'] = render
            return engine
        finally:
            self.lock.release()

    def preprocess_template(self, runtime_engine, name, ctx):
        self.lock.acquire(1)
        try:
            if name not in runtime_engine.renders:
                source = self.engine.render(name, ctx, {}, {})
                loader = runtime_engine.loader.loaders[0]
                loader.templates[name] = source
                runtime_engine.compile_template(name)
                del loader.templates[name]
        finally:
            self.lock.release()




"""
"""

__version__ = '0.1.132'

# flake8: noqa

from wheezy.template.engine import Engine
from wheezy.template.ext.code import CodeExtension
from wheezy.template.ext.core import CoreExtension
from wheezy.template.loader import DictLoader
from wheezy.template.loader import FileLoader
from wheezy.template.loader import PreprocessLoader




"""
"""

import imp

from wheezy.template.comp import adjust_source_lineno


class Compiler(object):

    def __init__(self, global_vars, source_lineno):
        self.global_vars = global_vars
        self.source_lineno = source_lineno

    def compile_module(self, source, name):
        source = adjust_source_lineno(source, name, self.source_lineno)
        compiled = compile(source, name, 'exec')
        module = imp.new_module(name)
        module.__name__ = name
        module.__dict__.update(self.global_vars)
        exec(compiled, module.__dict__)
        return module

    def compile_source(self, source, name):
        source = adjust_source_lineno(source, name, self.source_lineno)
        compiled = compile(source, name, 'exec')
        local_vars = {}
        exec(compiled, self.global_vars, local_vars)
        return local_vars




"""
"""


def parser_scan(extensions):
    parser_rules = {}
    parser_configs = []
    for extension in extensions:
        if hasattr(extension, 'parser_rules'):
            parser_rules.update(extension.parser_rules)
        if hasattr(extension, 'parser_configs'):
            parser_configs.extend(extension.parser_configs)
    return {
        'parser_rules': parser_rules,
        'parser_configs': parser_configs,
    }


class Parser(object):
    """
        ``continue_tokens`` are used to insert ``end`` node right
        before them to simulate a block end. Such nodes have token
        value ``None``.

        ``out_tokens`` are combined together into a single node.
    """

    def __init__(self, parser_rules, parser_configs=None, **ignore):
        self.end_tokens = []
        self.continue_tokens = []
        self.compound_tokens = []
        self.out_tokens = []
        self.rules = parser_rules
        if parser_configs:
            for config in parser_configs:
                config(self)

    def end_continue(self, tokens):
        """ If token is in ``continue_tokens`` prepend it
            with end token so it simulate a closed block.
        """
        for t in tokens:
            if t[1] in self.continue_tokens:
                yield (t[0], 'end', None)
            yield t

    def parse_iter(self, tokens):
        operands = []
        for lineno, token, value in tokens:
            if token in self.rules:
                value = self.rules[token](value)
            if token in self.out_tokens:
                operands.append((lineno, token, value))
            else:
                if operands:
                    yield operands[0][0], 'out', operands
                    operands = []
                if token in self.compound_tokens:
                    yield lineno, token, (
                        value, list(self.parse_iter(tokens)))
                else:
                    yield lineno, token, value
                    if token in self.end_tokens:
                        break
        if operands:
            yield operands[0][0], 'out', operands

    def parse(self, tokens):
        return list(self.parse_iter(self.end_continue(tokens)))




"""
"""

import re

from wheezy.template.comp import PY3
from wheezy.template.utils import find_all_balanced


# region: config

end_tokens = ['end']
continue_tokens = ['else:', 'elif ']
compound_tokens = ['for ', 'if ', 'def ', 'extends'] + continue_tokens
reserved_tokens = ['require', '#', 'include', 'import ', 'from ']
all_tokens = end_tokens + compound_tokens + reserved_tokens
out_tokens = ['markup', 'var', 'include']
known_var_filters = {
    's': PY3 and 'str' or 'unicode'
}


WRITER_DECLARE = '_b = []; w = _b.append'
WRITER_RETURN = "return ''.join(_b)"


# region: lexer extensions

def stmt_token(m):
    """ Produces statement token.
    """
    return m.end(), str(m.group(2)), str(m.group(1)).replace('\\\n', '')


RE_VAR = re.compile('(\.\w+)+')
RE_VAR_FILTER = re.compile('(?<!!)!\w+(!\w+)*')


def var_token(m):
    """ Produces variable token.
    """
    start = m.start(1)
    pos = m.end()
    source = m.string
    while 1:
        end = find_all_balanced(source, pos)
        if pos == end:
            break
        m = RE_VAR.match(source, end)
        if not m:
            break
        pos = m.end()
    value = str(source[start:end])
    m = RE_VAR_FILTER.match(source, end)
    if m:
        end = m.end()
        value += '!' + m.group()
    return end, 'var', value


# region: parser config

def configure_parser(parser):
    parser.end_tokens.extend(end_tokens)
    parser.continue_tokens.extend(continue_tokens)
    parser.compound_tokens.extend(compound_tokens)
    parser.out_tokens.extend(out_tokens)


# region: parser

def parse_require(value):
    return [v.strip(' ') for v in value.rstrip()[8:-1].split(',')]


def parse_extends(value):
    return value.rstrip()[8:-1]


def parse_include(value):
    return value.rstrip()[8:-1]


def parse_import(value):
    name, var = value[7:].rsplit(' as ', 1)
    return name, var


def parse_from(value):
    name, var = value[5:].rsplit(' import ', 1)
    s = var.rsplit(' as ', 1)
    if len(s) == 2:
        var, alias = s
    else:
        alias = var
    return name, var, alias


def parse_var(value):
    if '!!' not in value:
        return value, None
    var, var_filter = value.rsplit('!!', 1)
    return var, var_filter.split('!')


# region: block_builders

def build_extends(builder, lineno, token, nodes):
    assert token == 'render'
    if len(nodes) != 1:
        return False
    lineno, token, value = nodes[0]
    if token != 'extends':
        return False
    extends, nodes = value
    for lineno, token, value in nodes:
        if token in ('def ', 'require'):
            builder.build_token(lineno, token, value)
    lineno = builder.lineno
    builder.add(lineno + 1, 'return _r(' + extends +
                ', ctx, local_defs, super_defs)')
    return True


def build_module(builder, lineno, token, nodes):
    assert token == 'module'
    for lineno, token, value in nodes:
        if token == 'def ':
            builder.build_token(lineno, token, value)
    return True


def build_import(builder, lineno, token, value):
    assert token == 'import '
    name, var = value
    builder.add(lineno, var + ' = ' + '_i(' + name + ')')
    return True


def build_from(builder, lineno, token, value):
    assert token == 'from '
    name, var, alias = value
    builder.add(lineno, alias + ' = ' + '_i(' + name
                + ').local_defs[\'' + var + '\']')
    return True


def build_render_single_markup(builder, lineno, token, nodes):
    assert lineno <= 0
    assert token == 'render'
    if len(nodes) > 1:
        return False
    ln, token, nodes = nodes[0]
    if token != 'out' or len(nodes) > 1:
        return False
    ln, token, value = nodes[0]
    if token != 'markup':
        return False
    builder.add(ln, "return " + value)
    return True


def build_render(builder, lineno, token, nodes):
    assert lineno <= 0
    assert token == 'render'
    builder.add(lineno, WRITER_DECLARE)
    builder.build_block(nodes)
    lineno = builder.lineno
    builder.add(lineno + 1, WRITER_RETURN)
    return True


def build_def_syntax_check(builder, lineno, token, value):
    assert token == 'def '
    stmt, nodes = value
    lineno, token, value = nodes[0]
    if token in compound_tokens:
        builder.add(lineno, stmt)
        builder.start_block()
        token = token.rstrip()
        error = """\
The compound statement '%s' is not allowed here. \
Add a line before it with @#ignore.

%s
    @#ignore
    @%s ...""" % (token, stmt, token)
        builder.add(lineno, 'raise SyntaxError(%s)' % repr(error))
        builder.end_block()
        return True
    elif len(nodes) > 1:
        # token before @end
        lineno, token, value = nodes[-2]
        if token == '#':
            del nodes[-2]
    return False


def build_def_empty(builder, lineno, token, value):
    assert token == 'def '
    stmt, nodes = value
    if len(nodes) > 1:
        return False
    def_name = stmt[4:stmt.index('(', 5)]
    builder.add(lineno, stmt)
    builder.start_block()
    builder.add(lineno, "return ''")
    builder.end_block()
    builder.add(lineno + 1, def_name.join([
        "super_defs['", "'] = ", "; ",
        " = local_defs.setdefault('", "', ", ")"
    ]))
    return True


def build_def_single_markup(builder, lineno, token, value):
    assert token == 'def '
    stmt, nodes = value
    if len(nodes) > 2:
        return False
    ln, token, nodes = nodes[0]
    if token != 'out' or len(nodes) > 1:
        return False
    ln, token, value = nodes[0]
    if token != 'markup':
        return False
    def_name = stmt[4:stmt.index('(', 5)]
    builder.add(lineno, stmt)
    builder.start_block()
    builder.add(ln, "return " + value)
    builder.end_block()
    builder.add(ln + 1, def_name.join([
        "super_defs['", "'] = ", "; ",
        " = local_defs.setdefault('", "', ", ")"
    ]))
    return True


def build_def(builder, lineno, token, value):
    assert token == 'def '
    stmt, nodes = value
    def_name = stmt[4:stmt.index('(', 5)]
    builder.add(lineno, stmt)
    builder.start_block()
    builder.add(lineno + 1, WRITER_DECLARE)
    builder.build_block(nodes)
    lineno = builder.lineno
    builder.add(lineno, WRITER_RETURN)
    builder.end_block()
    builder.add(lineno + 1, def_name.join([
        "super_defs['", "'] = ", "; ",
        " = local_defs.setdefault('", "', ", ")"
    ]))
    return True


def build_out(builder, lineno, token, nodes):
    assert token == 'out'
    for lineno, token, value in nodes:
        if token == 'include':
            builder.add(lineno, 'w(' + '_r(' + value +
                        ', ctx, local_defs, super_defs)' + ')')
        elif token == 'var':
            var, var_filters = value
            if var_filters:
                for f in var_filters:
                    var = known_var_filters.get(f, f) + '(' + var + ')'
            builder.add(lineno, 'w(' + var + ')')
        elif value:
            builder.add(lineno, 'w(' + value + ')')
    return True


def build_compound(builder, lineno, token, value):
    assert token in compound_tokens
    stmt, nodes = value
    builder.add(lineno, stmt)
    builder.start_block()
    builder.build_block(nodes)
    builder.end_block()
    return True


def build_require(builder, lineno, token, variables):
    builder.add(lineno, '; '.join([
                name + " = ctx['" + name + "']" for name in variables]))
    return True


def build_comment(builder, lineno, token, comment):
    builder.add(lineno, comment)
    return True


def build_end(builder, lineno, token, value):
    if builder.lineno != lineno:
        builder.add(lineno - 1, '')
    return True


# region: core extension

class CoreExtension(object):
    """ Includes basic statements, variables processing and markup.
    """

    def __init__(self, token_start='@', line_join='\\'):

        def markup_token(m):
            """ Produces markup token.
            """
            return m.end(), 'markup', m.group().replace(
                token_start + token_start, token_start)

        self.lexer_rules = {
            100: (re.compile(r'%s((%s).*?(?<!\\))(\n|$)'
                             % (token_start, '|'.join(all_tokens)), re.S),
                  stmt_token),
            200: (re.compile(r'%s(\w+(\.\w+)*)' % token_start),
                  var_token),
            999: (re.compile(r'.+?(?=(?<!%s)%s(?!%s))|.+'
                             % (token_start, token_start, token_start),
                             re.S),
                  markup_token),
        }

        # region: preprocessors

        RE_CLEAN1 = re.compile(r'^([ ]+)%s(%s)'
                               % (token_start, '|'.join(all_tokens)), re.S)
        RE_CLEAN2 = re.compile(r'\n([ ]+)%s(%s)'
                               % (token_start, '|'.join(all_tokens)), re.S)

        def clean_source(source):
            """ Cleans leading whitespace before token start for all control
                tokens. Ignores escaped token start.
            """
            return RE_CLEAN2.sub(r'\n%s\2' % token_start, RE_CLEAN1.sub(
                r'%s\2' % token_start, source.replace('\r\n', '\n')))

        self.preprocessors = [clean_source]

        # region: parser

        if line_join:
            line_join += '\n'

            def parse_markup(value):
                value = value.replace(line_join, '')
                if value:
                    return repr(value)
                else:
                    return None
        else:
            def parse_markup(value):  # noqa
                if value:
                    return repr(value)
                else:
                    return None

        self.parser_rules = {
            'require': parse_require,
            'extends': parse_extends,
            'include': parse_include,
            'import ': parse_import,
            'from ': parse_from,
            'var': parse_var,
            'markup': parse_markup,
        }

    parser_configs = [configure_parser]

    builder_rules = [
        ('render', build_extends),
        ('render', build_render_single_markup),
        ('render', build_render),
        ('module', build_module),
        ('import ', build_import),
        ('from ', build_from),
        ('require', build_require),
        ('out', build_out),
        ('def ', build_def_syntax_check),
        ('def ', build_def_empty),
        ('def ', build_def_single_markup),
        ('def ', build_def),
        ('if ', build_compound),
        ('elif ', build_compound),
        ('else:', build_compound),
        ('for ', build_compound),
        ('#', build_comment),
        ('end', build_end),
    ]




"""
"""

import re

from wheezy.template.utils import find_balanced


# region: lexer extensions

def code_token(m):
    source = m.string
    start = m.end()
    end = find_balanced(source, start)
    if source[end::1] == '\n':
        end += 1
    return end, 'code', source[start:end]


# region: parser

def parse_code(value):
    lines = value.rstrip('\n')[1:-1].split('\n')
    lines[0] = lines[0].lstrip()
    if len(lines) == 1:
        return lines
    line = lines[1]
    n = len(line) - len(line.lstrip())
    return [l[:n].lstrip() + l[n:] for l in lines]


# region: block_builders

def build_code(builder, lineno, token, lines):
    for line in lines:
        builder.add(lineno, line)
        lineno += 1
    return True


# region: core extension

class CodeExtension(object):
    """ Includes support for embedded python code.
    """

    def __init__(self, token_start='@'):

        self.lexer_rules = {
            300: (re.compile(r'\s*%s(?=\()' % token_start), code_token),
        }

    parser_rules = {
        'code': parse_code
    }

    builder_rules = [
        ('code', build_code)
    ]




"""
"""

import os
import os.path
import stat

from time import time


class FileLoader(object):
    """ Loads templates from file system.

        ``directories`` - search path of directories to scan for template.
        ``encoding`` - decode template content per encoding.
    """

    def __init__(self, directories, encoding='UTF-8'):
        searchpath = []
        for path in directories:
            abspath = os.path.abspath(path)
            assert os.path.exists(abspath)
            assert os.path.isdir(abspath)
            searchpath.append(abspath)
        self.searchpath = searchpath
        self.encoding = encoding

    def list_names(self):
        """ Return a list of names relative to directories. Ignores any files
            and directories that start with dot.
        """
        names = []
        for path in self.searchpath:
            pathlen = len(path) + 1
            for dirpath, dirnames, filenames in os.walk(path):
                for i in [i for i, name in enumerate(dirnames)
                          if name.startswith('.')]:
                    del dirnames[i]
                for filename in filenames:
                    if filename.startswith('.'):
                        continue
                    name = os.path.join(dirpath, filename)[pathlen:]
                    name = name.replace('\\', '/')
                    names.append(name)
        return tuple(sorted(names))

    def get_fullname(self, name):
        """ Returns a full path by a template name.
        """
        for path in self.searchpath:
            filename = os.path.join(path, name)
            if not os.path.exists(filename):
                continue
            if not os.path.isfile(filename):
                continue
            return filename
        else:
            None

    def load(self, name):
        """ Loads a template by name from file system.
        """
        filename = self.get_fullname(name)
        if filename:
            f = open(filename, 'rb')
            try:
                return f.read().decode(self.encoding)
            finally:
                f.close()
        return None


class DictLoader(object):
    """ Loads templates from python dictionary.

        ``templates`` - a dict where key corresponds to template name and
        value to template content.
    """

    def __init__(self, templates):
        self.templates = templates

    def list_names(self):
        """ List all keys from internal dict.
        """
        return tuple(sorted(self.templates.keys()))

    def load(self, name):
        """ Returns template by name.
        """
        if name not in self.templates:
            return None
        return self.templates[name]


class ChainLoader(object):
    """ Loads templates from ``loaders`` until first succeed.
    """

    def __init__(self, loaders):
        self.loaders = loaders

    def list_names(self):
        """ Returns as list of names from all loaders.
        """
        names = set()
        for loader in self.loaders:
            names |= set(loader.list_names())
        return tuple(sorted(names))

    def load(self, name):
        """ Returns template by name from the first loader that succeed.
        """
        for loader in self.loaders:
            source = loader.load(name)
            if source:
                return source
        return None


class PreprocessLoader(object):
    """ Performs preprocessing of loaded template.
    """

    def __init__(self, engine, ctx=None):
        self.engine = engine
        self.ctx = ctx or {}

    def list_names(self):
        return self.engine.loader.list_names()

    def load(self, name):
        return self.engine.render(name, self.ctx, {}, {})


def autoreload(engine, enabled=True):
    """ Auto reload template if changes are detected in file. Limitation:
        inherited and imported templates.
    """
    if not enabled:
        return engine
    return AutoReloadProxy(engine)


# region: internal details

class AutoReloadProxy(object):

    def __init__(self, engine):
        self.engine = engine
        self.names = {}

    def get_template(self, name):
        if self.file_changed(name):
            self.remove(name)
        return self.engine.get_template(name)

    def render(self, name, ctx, local_defs, super_defs):
        if self.file_changed(name):
            self.remove(name)
        return self.engine.render(name, ctx, local_defs, super_defs)

    def remove(self, name):
        self.engine.remove(name)

    # region: internal details

    def __getattr__(self, name):
        return getattr(self.engine, name)

    def file_changed(self, name):
        try:
            last_known_stamp = self.names[name]
            current_time = int(time())
            if current_time - last_known_stamp <= 2:
                return False
        except KeyError:
            last_known_stamp = 0

        abspath = self.engine.loader.get_fullname(name)
        if not abspath:
            return False

        last_modified_stamp = os.stat(abspath)[stat.ST_MTIME]
        if last_modified_stamp <= last_known_stamp:
            return False
        self.names[name] = last_modified_stamp
        return True




"""
"""


def find_all_balanced(text, start=0):
    """ Finds balanced ``([`` with ``])`` assuming
        that ``start`` is pointing to ``(`` or ``[`` in ``text``.
    """
    if start >= len(text) or text[start] not in '([':
        return start
    while(1):
        pos = find_balanced(text, start)
        pos = find_balanced(text, pos, '[', ']')
        if pos != start:
            start = pos
        else:
            return pos


def find_balanced(text, start=0, start_sep='(', end_sep=')'):
    """ Finds balanced ``start_sep`` with ``end_sep`` assuming
        that ``start`` is pointing to ``start_sep`` in ``text``.
    """
    if start >= len(text) or start_sep != text[start]:
        return start
    balanced = 1
    pos = start + 1
    while pos < len(text):
        token = text[pos]
        pos += 1
        if token == end_sep:
            if balanced == 1:
                return pos
            balanced -= 1
        elif token == start_sep:
            balanced += 1
    return start


def print_source(source, lineno=1):  # pragma: nocover
    lines = []
    for line in source.split('\n'):
        lines.append("%02d " % lineno + line)
        lineno += line.count('\n') + 1
    print('\n'.join(lines))




"""
"""

import re

from wheezy.template.utils import find_balanced


RE_ARGS = re.compile(
    r'\s*(?P<expr>(([\'"]).*?\3|.+?))\s*\,')
RE_KWARGS = re.compile(
    r'\s*(?P<name>\w+)\s*=\s*(?P<expr>([\'"].*?[\'"]|.+?))\s*\,')
RE_STR_VALUE = re.compile(
    r'^[\'"](?P<value>.+)[\'"]$')
RE_INT_VALUE = re.compile(
    r'^(?P<value>(\d+))$')


# region: core extension

class DeterminedExtension(object):
    """ Tranlates function calls between template engines.

        Strictly determined known calls are converted to preprocessor
        calls, e.g.::

            @_('Name:')
            @path_for('default')
            @path_for('static', path='/static/css/site.css')

        Those that are not strictly determined are ignored and processed
        by runtime engine.
    """

    def __init__(self, known_calls, runtime_token_start='@',
                 token_start='#'):
        self.token_start = token_start
        self.pattern = re.compile(r'%s(%s)(?=\()' % (
            runtime_token_start, '|'.join(known_calls)))
        self.preprocessors = [self.preprocess]

    def preprocess(self, source):
        result = []
        start = 0
        for m in self.pattern.finditer(source):
            pstart = m.end()
            pend = find_balanced(source, pstart)
            if determined(source[pstart + 1:pend - 1]):
                name = m.group(1)
                result.append(source[start:m.start()])
                result.append(self.token_start + "ctx['" + name + "']")
                start = pstart
        if start:
            result.append(source[start:])
            return ''.join(result)
        else:
            return source


def determined(expression):
    """ Checks if expresion is strictly determined.

        >>> determined("'default'")
        True
        >>> determined('name')
        False
        >>> determined("'default', id=id")
        False
        >>> determined("'default', lang=100")
        True
        >>> determined('')
        True
    """
    args, kwargs = parse_params(expression)
    for arg in args:
        if not str_or_int(arg):
            return False
    for arg in kwargs.values():
        if not str_or_int(arg):
            return False
    return True


def parse_kwargs(text):
    """ Parses key-value type of parameters.

        >>> parse_kwargs('id=item.id')
        {'id': 'item.id'}
        >>> sorted(parse_kwargs('lang="en", id=12').items())
        [('id', '12'), ('lang', '"en"')]
    """
    kwargs = {}
    for m in RE_KWARGS.finditer(text + ','):
        groups = m.groupdict()
        kwargs[groups['name'].rstrip('_')] = groups['expr']
    return kwargs


def parse_args(text):
    """ Parses argument type of parameters.

        >>> parse_args('')
        []
        >>> parse_args('10, "x"')
        ['10', '"x"']
        >>> parse_args("'x', 100")
        ["'x'", '100']
        >>> parse_args('"default"')
        ['"default"']
    """
    args = []
    for m in RE_ARGS.finditer(text + ','):
        args.append(m.group('expr'))
    return args


def parse_params(text):
    """ Parses function parameters.

        >>> parse_params('')
        ([], {})
        >>> parse_params('id=item.id')
        ([], {'id': 'item.id'})
        >>> parse_params('"default"')
        (['"default"'], {})
        >>> parse_params('"default", lang="en"')
        (['"default"'], {'lang': '"en"'})
    """
    if '=' in text:
        args = text.split('=')[0]
        if ',' in args:
            args = args.rsplit(',', 1)[0]
            kwargs = text[len(args):]
            return parse_args(args), parse_kwargs(kwargs)
        else:
            return [], parse_kwargs(text)
    else:
        return parse_args(text), {}


def str_or_int(text):
    """ Ensures ``text`` as string or int expression.

        >>> str_or_int('"default"')
        True
        >>> str_or_int("'Hello'")
        True
        >>> str_or_int('100')
        True
        >>> str_or_int('item.id')
        False
    """
    m = RE_STR_VALUE.match(text)
    if m:
        return True
    else:
        m = RE_INT_VALUE.match(text)
        if m:
            return True
    return False




"""
"""


def builder_scan(extensions):
    builder_rules = {}
    for extension in extensions:
        if hasattr(extension, 'builder_rules'):
            rules = extension.builder_rules
            for token, builder in rules:
                builder_rules.setdefault(token, []).append(builder)
    return {
        'builder_rules': builder_rules
    }


class BlockBuilder(object):

    __slots__ = ('rules', 'indent', 'lineno', 'buf')

    def __init__(self, rules, indent='', lineno=0):
        self.rules = rules
        self.indent = indent
        self.lineno = lineno
        self.buf = []

    def start_block(self):
        self.indent += '    '

    def end_block(self):
        if len(self.indent) < 4:
            raise SyntaxError('Unexpected end of block.')
        self.indent = self.indent[:-4]

    def add(self, lineno, code):
        if lineno < self.lineno:
            raise SyntaxError('Inconsistence at %s : %s' %
                             (self.lineno, lineno))
        if lineno == self.lineno:
            line = self.buf[-1]
            if code:
                if line:
                    if line[-1:] == ':':
                        self.buf[-1] = line + code
                    else:
                        self.buf[-1] = line + '; ' + code
                else:
                    self.buf[-1] = self.indent + code
        else:
            pad = lineno - self.lineno - 1
            if pad > 0:
                self.buf.extend([''] * pad)
            if code:
                self.buf.append(self.indent + code)
            else:
                self.buf.append('')
        self.lineno = lineno + code.count('\n')

    def build_block(self, nodes):
        for lineno, token, value in nodes:
            self.build_token(lineno, token, value)

    def build_token(self, lineno, token, value):
        if token in self.rules:
            for rule in self.rules[token]:
                if rule(self, lineno, token, value):
                    break
        else:
            raise SyntaxError('No rule to build "%s" token at line %d.' %
                             (token, lineno))

    def to_string(self):
        return '\n'.join(self.buf)


class SourceBuilder(object):

    __slots__ = ('rules', 'lineno')

    def __init__(self, builder_rules, builder_offset=2, **ignore):
        self.rules = builder_rules
        self.lineno = 0 - builder_offset

    def build_source(self, nodes):
        builder = BlockBuilder(self.rules)
        builder.build_block(nodes)
        return builder.to_string()

    def build_render(self, nodes):
        builder = BlockBuilder(self.rules, lineno=self.lineno)
        builder.add(self.lineno + 1,
                    'def render(ctx, local_defs, super_defs):')
        builder.start_block()
        builder.build_token(self.lineno + 2, 'render', nodes)
        return builder.to_string()

    def build_module(self, nodes):
        builder = BlockBuilder(self.rules, lineno=self.lineno)
        builder.add(self.lineno + 1, 'local_defs = {}; super_defs = {}')
        builder.build_token(self.lineno + 2, 'module', nodes)
        return builder.to_string()




"""
"""

from wheezy.template.builder import SourceBuilder
from wheezy.template.builder import builder_scan
from wheezy.template.comp import allocate_lock
from wheezy.template.compiler import Compiler
from wheezy.template.lexer import Lexer
from wheezy.template.lexer import lexer_scan
from wheezy.template.parser import Parser
from wheezy.template.parser import parser_scan


class Engine(object):
    """ The core component of template engine.
    """

    def __init__(self, loader, extensions, template_class=None):
        self.lock = allocate_lock()
        self.templates = {}
        self.renders = {}
        self.modules = {}
        self.global_vars = {
            '_r': self.render,
            '_i': self.import_name
        }
        self.loader = loader
        self.template_class = template_class or Template
        self.compiler = Compiler(self.global_vars, -2)
        self.lexer = Lexer(**lexer_scan(extensions))
        self.parser = Parser(**parser_scan(extensions))
        self.builder = SourceBuilder(**builder_scan(extensions))

    def get_template(self, name):
        """ Returns compiled template.
        """
        try:
            return self.templates[name]
        except KeyError:
            self.compile_template(name)
            return self.templates[name]

    def render(self, name, ctx, local_defs, super_defs):
        """ Renders template by name in given context.
        """
        try:
            return self.renders[name](ctx, local_defs, super_defs)
        except KeyError:
            self.compile_template(name)
            return self.renders[name](ctx, local_defs, super_defs)

    def remove(self, name):
        """ Removes given ``name`` from internal cache.
        """
        self.lock.acquire(1)
        try:
            if name in self.renders:
                del self.templates[name]
                del self.renders[name]
            if name in self.modules:
                del self.modules[name]
        finally:
            self.lock.release()

    # region: internal details

    def import_name(self, name):
        try:
            return self.modules[name]
        except KeyError:
            self.compile_import(name)
            return self.modules[name]

    def compile_template(self, name):
        self.lock.acquire(1)
        try:
            if name not in self.renders:
                template_source = self.loader.load(name)
                if template_source is None:
                    raise IOError('Template "%s" not found.' % name)
                tokens = self.lexer.tokenize(template_source)
                nodes = self.parser.parse(tokens)
                source = self.builder.build_render(nodes)

                #self.print_debug(name, tokens, nodes, source)

                render_template = self.compiler.compile_source(
                    source, name)['render']
                self.renders[name] = render_template
                self.templates[name] = self.template_class(
                    name, render_template)
        finally:
            self.lock.release()

    def compile_import(self, name):
        self.lock.acquire(1)
        try:
            if name not in self.modules:
                template_source = self.loader.load(name)
                if template_source is None:
                    raise IOError('Import "%s" not found.' % name)
                tokens = self.lexer.tokenize(template_source)
                nodes = self.parser.parse(tokens)
                source = self.builder.build_module(nodes)

                #self.print_debug(name, tokens, nodes, source)

                self.modules[name] = self.compiler.compile_module(
                    source, name)
        finally:
            self.lock.release()

    def print_debug(self, name, tokens, nodes, source):  # pragma: nocover
        print(name.center(80, '-'))
        from pprint import pprint
        #pprint(tokens)
        pprint(nodes)
        from wheezy.template.utils import print_source
        print_source(source, -1)


class Template(object):
    """ Simple template class.
    """

    def __init__(self, name, render_template):
        self.name = name
        self.render_template = render_template

    def render(self, ctx):
        return self.render_template(ctx, {}, {})




""" ``comp`` module.
"""

import sys


PY3 = sys.version_info[0] >= 3


if PY3:  # pragma: nocover
    from _thread import allocate_lock
else:  # pragma: nocover
    from thread import allocate_lock  # noqa


try:  # pragma: nocover
    import ast

    def adjust_source_lineno(source, name, lineno):
        source = compile(source, name, 'exec', ast.PyCF_ONLY_AST)
        ast.increment_lineno(source, lineno)
        return source

except ImportError:  # pragma: nocover

    def adjust_source_lineno(source, name, lineno):  # noqa
        return source
